/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#ifndef __ATOMIC_ATOMIC_H
#define __ATOMIC_ATOMIC_H

#include <atomic/target.h>
#include <atomic/glue.h>

#define mc()	barrier()
#define rmc()	barrier()
#define wmc()	barrier()

#define smp_mb()	mb()
#define smp_rmb()	rmb()
#define smp_wmb()	wmb()
#define smp_mc()	mc()
#define smp_rmc()	rmc()
#define smp_wmc()	wmc()

#if defined TARGET_ARCH_I386 || defined TARGET_ARCH_X86_64
#include <atomic/arch_x86.h>
#elif defined TARGET_ARCH_PPC
#include <atomic/arch_ppc.h>
#elif defined TARGET_ARCH_SPARC
#include <atomic/arch_sparc.h>
#elif defined HAVE_SYNC_ATOMIC_SUPPORT
typedef struct {
	volatile int		val;
} atomic_t;

static inline void atomic_set(atomic_t *a, int val)
{
	a->val = val;
}

static inline int atomic_inc(atomic_t *a)
{
	return __sync_add_and_fetch(&a->val, 1);
}

static inline int atomic_dec(atomic_t *a)
{
	return __sync_sub_and_fetch(&a->val, 1);
}

#define atomic_dec_and_test(a) (atomic_dec(a) == 0)

#else
#error "Target arch is not supported." 
#endif

#endif /* __ATOMIC_ATOMIC_H */
